# -*- coding: utf-8 -*-
'''
Control and add subsystems to the running daemon hub
'''
# Import python libs
import os
# Import pop libs
import pop.hub


def add(hub,
        pypath=None,
        subname=None,
        sub=None,
        static=None,
        contracts_pypath=None,
        contracts_static=None,
        default_contracts=None,
        virtual=True,
        dyne_name=None,
        omit_start=('_'),
        omit_end=(),
        omit_func=False,
        omit_class=True,
        omit_vars=False,
        mod_basename='pop.sub',
        stop_on_failures=False,
        init=True,
        ):
    '''
    Add a new subsystem to the hub
    '''
    if pypath:
        pypath = pop.hub.ex_path(pypath)
        subname = subname if subname else pypath[0].split('.')[-1]
    elif static:
        subname = subname if subname else os.path.basename(static)
    if dyne_name:
        subname = subname if subname else dyne_name
    if sub:
        root = sub
    else:
        root = hub
    root._subs[subname] = pop.hub.Sub(
            hub,
            subname,
            pypath,
            static,
            contracts_pypath,
            contracts_static,
            default_contracts,
            virtual,
            dyne_name,
            omit_start,
            omit_end,
            omit_func,
            omit_class,
            omit_vars,
            mod_basename,
            stop_on_failures)
    root._subs[subname]._sub_init(init)
    root._iter_subs = sorted(root._subs.keys())


def remove(hub, subname):
    '''
    Remove a pop from the hub, run the shutdown if needed
    '''
    if hasattr(hub, subname):
        sub = getattr(hub, subname)
        if hasattr(sub, 'init'):
            mod = getattr(sub, 'init')
            if hasattr(mod, 'shutdown'):
                mod.shutdown()
        hub._remove_subsystem(subname)


def load_all(hub, subname):
    '''
    Load all modules under a given pop
    '''
    if hasattr(hub, subname):
        sub = getattr(hub, subname)
        sub._load_all()
        return True
    else:
        return False


def get_dirs(hub, sub):
    '''
    Return a list of directories that contain the modules for this subname
    '''
    return sub._dirs


def iter_subs(hub, sub):
    '''
    Return an iterator that will traverse just the subs. This is useful for
    nested subs
    '''
    for name in sorted(sub._subs):
        yield sub._subs[name]


def load_subdirs(hub, sub, recurse=False):
    '''
    Given a sub, load all subdirectories found under the sub into a lower namespace
    '''
    dirs = hub.pop.sub.get_dirs(sub)
    roots = {}
    for dir_ in dirs:
        for fn in os.listdir(dir_):
            if fn.startswith('_'):
                continue
            if fn == 'contracts':
                continue
            full = os.path.join(dir_, fn)
            if not os.path.isdir(full):
                continue
            if fn not in roots:
                roots[fn] = [full]
            else:
                roots[fn].append(full)
    for name, sub_dirs in roots.items():
        # Load er up!
        hub.pop.sub.add(
                subname=name,
                sub=sub,
                static=sub_dirs,
                virtual=sub._virtual,
                omit_start=sub._omit_start,
                omit_end=sub._omit_end,
                omit_func=sub._omit_func,
                omit_class=sub._omit_class,
                omit_vars=sub._omit_vars,
                mod_basename=sub._mod_basename,
                stop_on_failures=sub._stop_on_failures)
        if recurse:
            hub.pop.sub.load_subdirs(getattr(sub, name), recurse)


def reload(hub, subname):
    '''
    Instruct the hub to reload the modules for the given sub. This does not call
    the init.new function or remove sub level variables. But it does re-read the
    directory list and re-initialize the loader causing all modules to be re-evaluated
    when started.
    '''
    if hasattr(hub, subname):
        sub = getattr(hub, subname)
        sub._prepare()
        return True
    else:
        return False


def extend(
        hub,
        subname,
        pypath=None,
        static=None,
        contracts_pypath=None,
        contracts_static=None):
    '''
    Extend the directory lookup for a given sub. Any of the directory lookup
    arguments can be passed.
    '''
    if not hasattr(hub, subname):
        return False
    sub = getattr(hub, subname)
    if pypath:
        sub._pypath.extend(pop.hub.ex_path(pypath))
    if static:
        sub._static.extend(pop.hub.ex_path(static))
    if contracts_pypath:
        sub._contracts_pypath.extend(pop.hub.ex_path(contracts_pypath))
    if contracts_static:
        sub._contracts_static.extend(pop.hub.ex_path(contracts_static))
    sub._prepare()



'''
Seed a new project with a directory tree and first files
'''
# Import python libs
import os

BUILD = """#!/usr/bin/python3
import sys
import os
import shutil
import subprocess
import tempfile
import venv
import argparse

OMIT = ('__pycache__', 'PyInstaller', 'pip', 'setuptools', 'pkg_resources', '__pycache__', 'dist-info', 'egg-info')


def parse():
    '''
    Parse the cli args
    '''
    parser = argparse.ArgumentParser()
    parser.add_argument(
            'name',
            help='The name of the script to build from the run.py')
    args = parser.parse_args()
    return args.__dict__


class Builder:
    def __init__(self):
        self.opts = parse()
        self.name = self.opts['name']
        self.cwd = os.getcwd()
        self.run = os.path.join(self.cwd, 'run.py')
        self.venv_dir = tempfile.mkdtemp(prefix='pop_', suffix='_venv')
        self.python_bin = os.path.join(self.venv_dir, 'bin', 'python')
        self.vroot = os.path.join(self.venv_dir, 'lib')
        self.all_paths = set()
        self.imports = set()
        self.datas = set()
        self.cmd = f'{self.python_bin} -B -OO -m PyInstaller '
        self.s_path = os.path.join(self.venv_dir, 'bin', self.name)
        self.pyi_args = [
              self.s_path,
              '--log-level=INFO',
              '--noconfirm',
              '--onefile',
              '--clean',
            ]

    def create(self):
        '''
        Make a virtual environment based on the version of python used to call this script
        '''
        venv.create(self.venv_dir, clear=True, with_pip=True)
        pip_bin = os.path.join(self.venv_dir, 'bin', 'pip')
        subprocess.call([pip_bin, 'install', '-r', 'requirements.txt'])
        subprocess.call([pip_bin, 'install', 'PyInstaller'])
        subprocess.call([pip_bin, '-v', 'install', self.cwd])

    def omit(self, test):
        for bad in OMIT:
            if bad in test:
                return True
        return False

    def scan(self):
        '''
        Scan the new venv for files and imports
        '''
        for root, dirs, files in os.walk(self.vroot):
            if self.omit(root):
                continue
            for d in dirs:
                full = os.path.join(root, d)
                if self.omit(full):
                    continue
                self.all_paths.add(full)
            for f in files:
                full = os.path.join(root, f)
                if self.omit(full):
                    continue
                self.all_paths.add(full)

    def to_import(self, path): 
        ret = path[path.index('site-packages') + 14:].replace(os.sep, '.')
        if ret.endswith('.py'):
            ret = ret[:-3]
        return ret

    def to_data(self, path):
        dest = path[path.index('site-packages') + 14:]
        src = path
        if not dest.strip():
            return None
        ret = f'{src}{os.pathsep}{dest}'
        return ret

    def mk_adds(self):
        '''
        make the imports and datas for pyinstaller
        '''
        for path in self.all_paths:
            if not 'site-packages' in path:
                continue
            if os.path.isfile(path):
                if not path.endswith('.py'):
                    continue
                if path.endswith('__init__.py'):
                    # Skip it, we will get the dir
                    continue
                imp = self.to_import(path)
                if imp:
                    self.imports.add(imp)
            if os.path.isdir(path):
                data = self.to_data(path)
                imp = self.to_import(path)
                if imp:
                    self.imports.add(imp)
                if data:
                    self.datas.add(data)

    def mk_cmd(self):
        '''
        Create the pyinstaller command
        '''
        for imp in self.imports:
            self.pyi_args.append(f'--hidden-import={imp}')
        for data in self.datas:
            self.pyi_args.append(f'--add-data={data}')
        for arg in self.pyi_args:
            self.cmd += f'{arg} '

    def pyinst(self):
        shutil.copy(self.run, self.s_path)
        subprocess.call(self.cmd, shell=True)

    def report(self):
        art = os.path.join(self.cwd, 'dist', self.name)
        print(f'Executable created in {art}')
        print('To create a more portable and fully static binary install run staticx against your new build')

    def clean(self):
        shutil.rmtree(self.venv_dir)
        shutil.rmtree(os.path.join(self.cwd, 'build'))
        os.remove(os.path.join(self.cwd, f'{self.name}.spec'))

    def build(self):
        self.create()
        self.scan()
        self.mk_adds()
        self.mk_cmd()
        self.pyinst()
        self.report()
        self.clean()


if __name__ == '__main__':
    builder = Builder()
    builder.build()
"""

SETUP = '''#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Import python libs
import os
import sys
import shutil
from setuptools import setup, Command

NAME = '%%NAME%%'
DESC = ('')

# Version info -- read without importing
_locals = {}
with open('{}/version.py'.format(NAME)) as fp:
    exec(fp.read(), None, _locals)
VERSION = _locals['version']
SETUP_DIRNAME = os.path.dirname(__file__)
if not SETUP_DIRNAME:
    SETUP_DIRNAME = os.getcwd()


class Clean(Command):
    user_options = []
    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for subdir in (NAME, 'tests'):
            for root, dirs, files in os.walk(os.path.join(os.path.dirname(__file__), subdir)):
                for dir_ in dirs:
                    if dir_ == '__pycache__':
                        shutil.rmtree(os.path.join(root, dir_))


def discover_packages():
    modules = []
    for package in (NAME, ):
        for root, _, files in os.walk(os.path.join(SETUP_DIRNAME, package)):
            pdir = os.path.relpath(root, SETUP_DIRNAME)
            modname = pdir.replace(os.sep, '.')
            modules.append(modname)
    return modules


setup(name=NAME,
      author='',
      author_email='',
      url='',
      version=VERSION,
      description=DESC,
      classifiers=[
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Development Status :: 5 - Production/Stable',
          ],
      entry_points={
        'console_scripts': [
            '%%NAME%% = %%NAME%%.scripts:start',
            ],
          },
      packages=discover_packages(),
      cmdclass={'clean': Clean},
      )
'''

SCRIPT = '''#!/usr/bin/env python3
import pop.hub

def start():
    hub = pop.hub.Hub()
    hub.pop.sub.add('%%NAME%%.%%NAME%%')
'''

INIT = '''def __init__(hub):
    print('%%NAME%% works!')
'''

REQ = 'pop'

CONF = '''CLI_CONFIG = {}
CONFIG = {}
GLOBAL = {}
SUBS = {}
DYNE = {}
'''

VER = '''version = '1.0.0'
'''


def new(hub):
    '''
    Given the option in hub.opts "seed_name" create a directory tree for a
    new pop project
    '''
    hub.PATH = os.getcwd()
    name = hub.opts['seed_name']
    hub.pop.seed.mkdir(name, name)
    hub.pop.seed.mkdir(name, name, 'contracts')
    hub.pop.seed.mksetup(name)
    hub.pop.seed.mkscript(name)
    hub.pop.seed.mkrun(name)
    hub.pop.seed.mkinit(name)
    hub.pop.seed.mkversion(name)
    hub.pop.seed.mkconf(name)
    hub.pop.seed.mkreq(name)
    hub.pop.seed.mkbuild()


def mkdir(hub, *args):
    '''
    Create the named dir
    '''
    path = hub.PATH
    for dir_ in args:
        path = os.path.join(path, dir_)
        if not os.path.isdir(path):
            try:
                os.makedirs(path)
            except Exception:
                print('Failed to make {}'.format(path))
                continue
            if dir_ == 'scripts' and len(args) == 1:
                continue


def mkreq(hub, name):
    '''
    '''
    path = os.path.join(hub.PATH, 'requirements.txt')
    with open(path, 'w+') as fp:
        fp.write(REQ)


def mksetup(hub, name):
    '''
    Create and write out a setup.py file
    '''
    path = os.path.join(hub.PATH, 'setup.py')
    setup_str = SETUP.replace('%%NAME%%', name)
    with open(path, 'w+') as fp:
        fp.write(setup_str)


def mkscript(hub, name):
    '''
    Create and write out a setup.py file
    '''
    path = os.path.join(hub.PATH, name, 'scripts.py')
    script_str = SCRIPT.replace('%%NAME%%', name)
    with open(path, 'w+') as fp:
        fp.write(script_str)


def mkrun(hub, name):
    '''
    Create the convenience run.py script allowing the project to
    be executed from the local directory
    '''
    path = os.path.join(hub.PATH, 'run.py')
    run_str = SCRIPT.replace('%%NAME%%', name)
    run_str += '\n\nstart()'
    with open(path, 'w+') as fp:
        fp.write(run_str)


def mkinit(hub, name):
    '''
    Create the intial init.py
    '''
    path = os.path.join(hub.PATH, name, name, 'init.py')
    init_str = INIT.replace('%%NAME%%', name)
    with open(path, 'w+') as fp:
        fp.write(init_str)


def mkversion(hub, name):
    '''
    Create the version.py file
    '''
    path = os.path.join(hub.PATH, name, 'version.py')
    with open(path, 'w+') as fp:
        fp.write(VER)


def mkconf(hub, name):
    '''
    Create the version.py file
    '''
    path = os.path.join(hub.PATH, name, 'conf.py')
    with open(path, 'w+') as fp:
        fp.write(CONF)


def mkbuild(hub):
    '''
    Create the build script to make the single executable
    '''
    path = os.path.join(hub.PATH, 'build.py')
    with open(path, 'w+') as fp:
        fp.write(BUILD)



'''
Used to take care of the options that end in `_dir`. The assumption is that
`_dir` options need to be treated differently. They need to verified to exist
and they need to be rooted based on the user, root option etc.
'''
# Import python libs
import os


def roots(hub, default_root, opts, home_root):
    '''
    Detect the root dir data and apply it
    '''
    # TODO: Make this safe for Windows
    os_root = '/'
    root = os_root
    change = False
    anchor = ''
    if not os.geteuid() == 0:
        root = os.path.join(os.environ['HOME'], f'.{home_root}')
        change = True
        anchor = home_root
    if opts.get('root_dir', root) != default_root:
        root = opts.get('root_dir', root)
        change = True
    if not root.endswith(os.sep):
        root = f'{root}{os.sep}'
    if change:
        for key in opts:
            if key == 'root_dir':
                continue
            if key.endswith('_dir'):
                if anchor:
                    if anchor in opts[key]:
                        a_len = len(anchor) + 1
                        opts[key] = f'{os_root}{opts[key][opts[key].index(anchor)+a_len:]}'
                opts[key] = opts[key].replace(
                        os_root, root, 1)


def verify(hub, opts):
    '''
    Verify that the environment and all named directories in the
    configuration exist
    '''
    for key in opts:
        if key == 'root_dir':
            continue
        if key == 'config_dir':
            continue
        if key.endswith('_dir'):
            if not os.path.isdir(opts[key]):
                os.makedirs(opts[key])



# -*- coding: utf-8 -*-
'''
The main interface for management of the aio loop
'''
# Import python libs
import asyncio
import os
import sys
import signal
import functools

__virtualname__ = 'loop'


def __virtual__(hub):
    return True


def create(hub):
    '''
    Create the loop at hub.pop.Loop
    '''
    if not hub.pop.Loop:
        hub.pop.loop.FUT_QUE = asyncio.Queue()
        if sys.platform == 'win32':
            hub.pop.Loop = asyncio.ProactorEventLoop()
        else:
            hub.pop.Loop = asyncio.get_event_loop()


def call_soon(hub, ref, *args, **kwargs):
    '''
    Schedule a coroutine to be called when the loop has time. This needs
    to be called after the creation fo the loop
    '''
    fun = hub.pop.ref.get_func(ref)
    hub.pop.Loop.call_soon(functools.partial(fun, *args, **kwargs))


def ensure_future(hub, ref, *args, **kwargs):
    '''
    Schedule a coroutine to be called when the loop has time. This needs
    to be called after the creation fo the loop. This function also uses
    the hold system to await the future when it is done making it easy
    to create a future that will be cleanly awaited in the background.
    '''
    fun = getattr(hub, ref)
    future = asyncio.ensure_future(fun(*args, **kwargs))

    def callback(fut):
        hub.pop.loop.FUT_QUE.put_nowait(fut)
    future.add_done_callback(callback)
    return future


def start(hub, *coros, hold=False, sigint=None, sigterm=None):
    '''
    Start a loop that will run until complete
    '''
    hub.pop.loop.create()
    if sigint:
        s = signal.SIGINT
        hub.pop.Loop.add_signal_handler(
            s, lambda s=s: asyncio.create_task(sigint(s))
            )
    if sigterm:
        s = signal.SIGTERM
        hub.pop.Loop.add_signal_handler(
            s, lambda s=s: asyncio.create_task(sigterm(s))
            )
    if hold:
        coros = list(coros)
        coros.append(_holder(hub))
    # DO NOT CHANGE THIS CALL TO run_forever! If we do that then the tracebacks
    # do not get resolved.
    return hub.pop.Loop.run_until_complete(
            asyncio.gather(*coros)
            )


async def _holder(hub):
    '''
    Just a sleeping while loop to hold the loop open while it runs until
    complete
    '''
    while True:
        future = await hub.pop.loop.FUT_QUE.get()
        await future


async def await_futures(hub):
    '''
    Scan over the futures that have completed and manually await them.
    This function is used to clean up futures when the loop is not opened
    up with hold=True so that ensured futures can be cleaned up on demand
    '''
    while not hub.pop.loop.FUT_QUE.empty():
        future = await hub.pop.loop.FUT_QUE.get()
        await future


async def kill(hub, wait=0):
    '''
    Close out the loop
    '''
    await asyncio.sleep(wait)
    hub.pop.Loop.stop()
    while True:
        if not hub.pop.Loop.is_running():
            hub.pop.Loop.close()
        await asyncio.sleep(1)


async def as_yielded(hub, gens):
    '''
    Concurrently run multiple async generators and yield the next yielded
    value from the soonest yielded generator.

    async def many():
        for n in range(10):
            yield os.urandom(6).hex()

    async def run():
        gens = []
        for n in range(10):
            gens.append(many())
        async for y in as_yielded(gens):
            print(y)
    '''
    fin = os.urandom(32)
    que = asyncio.Queue()
    fs = []
    to_clean = []
    async def _yield(gen):
        async for comp in gen:
            await que.put(comp)
    async def _ensure(coros):
        for f in asyncio.as_completed(coros):
            await f
    async def _set_done():
        await que.put(fin)
    def _done(future):
        to_clean.append(asyncio.ensure_future(_set_done()))
    coros = []
    for gen in gens:
        coros.append(_yield(gen))
    f = asyncio.ensure_future(_ensure(coros))
    f.add_done_callback(_done)
    while True:
        ret = await que.get()
        if ret == fin:
            break
        yield ret
    for c in to_clean:
        await c



#!/usr/bin/python3

import pop.hub


def pop_seed():
    CONFIG = {
            'seed_name': {
                'positional': True,
                'help': 'The name of the project that is being created',
                },
            }

    hub = pop.hub.Hub()
    hub.pop.sub.add('pop.mods.conf')
    hub.opts = hub.conf.reader.read(CONFIG)
    hub.pop.seed.new()



# -*- coding: utf-8 -*-
'''
The reader module is used to read the config data. This will read in cli
arguments and merge them with config fie arguments.
'''

# Priority order: cli, config, cli_defaults

__virtualname__ = 'reader'
__contracts__ = [__virtualname__]


def _merge_dicts(opts, updates, os_opts, explicit_cli_args):
    '''
    recursively merge updates into opts
    '''
    for key, val in os_opts.items():
        if not val:
            # Don't use empty os vals
            continue
        if key in opts:
            opts[key] = val
    for key, val in updates.items():
        if isinstance(val, dict) and isinstance(opts.get(key), dict):
            _merge_dicts(opts.get(key, {}), val, os_opts, explicit_cli_args)
        elif val is not None:
            if key not in opts:
                # The key is not in opts(from config file), let's add it
                opts[key] = val
                continue

            # We already have a value for the key in opts
            if opts[key] == val:
                # The value is the same, carry on
                continue

            if key in explicit_cli_args:
                # We have a value for the key in opts(from config file) but
                # this option was explicitly passed on the CLI, ie, it's not
                # a default value.
                # Overwrite what's in opts
                opts[key] = val
                continue
    return opts


def read(hub,
        defaults,
        subs=None,
        loader='json',
        process_cli=True,
        process_cli_known_args_only=False,
        args=None,
        namespace=None):
    '''
    Pass in the default options dict to use
    :param opts:
    :param process_cli: Process the passed args or sys.argv
    :param process_cli_known_args_only: Tells the ArgumentParser to only process known arguments
    :param args: Arguments to pass to ArgumentParser
    :param namespace: argparse.Namespace to pass to ArgumentParser
    :return: options
    '''
    hub.conf._loader = loader
    if subs:
        hub.conf.args.subs(subs)
    opts = hub.conf.args.setup(defaults)['return']
    os_opts = hub.conf.os.gather(defaults)
    if process_cli is True:
        cli_opts = hub.conf.args.parse(args, namespace, process_cli_known_args_only)['return']
    else:
        cli_opts = {}
    explicit_cli_args = cli_opts.pop('_explicit_cli_args_', set())
    cli_opts = hub.conf.args.render(defaults, cli_opts, explicit_cli_args)
    kwargs = {}
    # Due to the order of priorities and the representation of defaults in the
    # Argparser we need to manually check if the config option values are from
    # the cli or from defaults
    f_func = False
    if 'config_dir' in cli_opts:
        if cli_opts['config_dir']:
            kwargs['confdir'] = cli_opts['config_dir']
        else:
            kwargs['confdir'] = opts['config_dir']
        if 'config_recurse' in cli_opts:
            if cli_opts['config_recurse']:
                kwargs['recurse'] = cli_opts['config_recurse']
            else:
                kwargs['recurse'] = opts['config_recurse']
        # If the config_dir configuration dictionary provides a configuration
        # file pattern to read, pass it along
        kwargs['pattern'] = defaults['config_dir'].get('pattern')
        f_func = hub.conf.file.load_dir
    elif 'config' in cli_opts:
        if cli_opts['config']:
            kwargs['paths'] = cli_opts['config']
        else:
            kwargs['paths'] = opts['config']
        f_func = hub.conf.file.load_file
    # Render args before config parsing
    if f_func:
        f_opts = f_func(**kwargs)
        opts.update(f_opts)
        return _merge_dicts(opts, cli_opts, os_opts, explicit_cli_args)
    else:
        return _merge_dicts(opts, cli_opts, os_opts, explicit_cli_args)



# Import python libs
import logging


def setup(hub, conf):
    '''
    Given the configuration data set up the logger
    '''
    level = hub.conf.log.LEVELS.get(conf['log_level'], logging.INFO)
    root = logging.getLogger('')
    root.setLevel(level)
    cf = logging.Formatter(fmt=conf['log_fmt_console'], datefmt=conf['log_datefmt'])
    ch = logging.StreamHandler()
    ch.setLevel(level)
    ch.setFormatter(cf)
    root.addHandler(ch)
    ff = logging.Formatter(fmt=conf['log_fmt_console'], datefmt=conf['log_datefmt'])
    fh = logging.FileHandler(conf['log_file'])
    fh.setLevel(level)
    fh.setFormatter(ff)
    root.addHandler(fh)



# -*- coding: utf-8 -*-
'''
Pop related exceptions
'''


class PopBaseException(Exception):
    '''
    Base exception where all of Pop's exceptions derive
    '''


class PopError(PopBaseException):
    '''
    General purpose pack exception to signal an error
    '''


class PopLoadError(PopBaseException):
    '''
    Exception raised when a pack module fails to load
    '''


class PopLookupError(PopBaseException):
    '''
    Exception raised when a pack module lookup fails
    '''


class ContractModuleException(PopBaseException):
    '''
    Exception raised when a function specified in a contract as required
    to exist is not found in the loaded module
    '''


class ContractFuncException(PopBaseException):
    '''
    Exception raised when a function specified in a contract as required
    to exist is found on the module but it's not function
    '''

class ContractSigException(PopBaseException):
    '''
    Exception raised when a function signature is not compatible with the
    coresponding function signature found in the contract. 
    '''


class ProcessNotStarted(PopBaseException):
    '''
    Exception raised when failing to start a process on the process manager
    '''


class BindError(PopBaseException):
    '''
    Exception raised when arguments for a function in a ContractedContext cannot be bound
    Indicates invalid function arguments.
    '''



'''
Convenience wrappers to make using the conf system as easy and seamless as possible
'''


def integrate(
        hub,
        imports,
        override=None,
        cli=None,
        roots=None,
        home_root=None,
        loader='json',
        logs=True):
    '''
    Load the conf sub and run the integrate sequence.
    '''
    hub.pop.sub.add('pop.mods.conf')
    hub.conf.integrate.load(
        imports,
        override,
        cli=cli,
        roots=roots,
        home_root=home_root,
        loader=loader,
        logs=logs)



# -*- coding: utf-8 -*-
'''
Used to scan the given directories for loadable files
'''
# Import python libs
import os
import importlib
import collections

PY_END = ('.py', '.pyc', '.pyo')
PYEXT_END = tuple(importlib.machinery.EXTENSION_SUFFIXES)
CYTHON_END = ('.pyx',)
SKIP_DIRNAMES = ('__pycache__',)


def scan(dirs):
    '''
    Return a list of importable files
    '''
    ret = collections.OrderedDict()
    ret['python'] = collections.OrderedDict()
    ret['cython'] = collections.OrderedDict()
    ret['ext'] = collections.OrderedDict()
    ret['imp'] = collections.OrderedDict()
    for dir_ in dirs:
        for fn_ in os.listdir(dir_):
            _apply_scan(ret, dir_, fn_)
    return ret


def _apply_scan(ret, dir_, fn_):
    '''
    Convert the scan data into
    '''
    if fn_.startswith('_'):
        return
    if os.path.basename(dir_) in SKIP_DIRNAMES:
        return
    full = os.path.join(dir_, fn_)
    if '.' not in full:
        return
    bname = full[:full.rindex('.')]
    if fn_.endswith(PY_END):
        if bname not in ret['python']:
            ret['python'][bname] = {'path': full}
    if fn_.endswith(CYTHON_END):
        if bname not in ret['cython']:
            ret['cython'][bname] = {'path': full}
    if fn_.endswith(PYEXT_END):
        if bname not in ret['ext']:
            ret['ext'][bname] = {'path': full}



# -*- coding: utf-8 -*-
'''
Define the yaml loader interface
'''

# Import third party libs
try:
    import yaml
    HAS_YAML = True
except ImportError:
    HAS_YAML = False

__virtualname__ = 'yaml'
__contracts__ = [__virtualname__]


def __virtual__(hub):
    if HAS_YAML:
        return True
    return (False, 'PyYaml could not be loaded')


def load(hub, path):
    '''
    use yaml to read in a file
    '''
    try:
        with open(path, 'rb') as fp_:
            return yaml.safe_load(fp_.read())
    except FileNotFoundError:
        pass
    return {}


def render(hub, val):
    '''
    Take the string and render it in json
    '''
    return yaml.safe_load(val)



'''
Integrate is used to pull config data from multiple sources and merge it into
the hub. Once it is merged then when a sub is loaded the respective config data
is loaded into the sub as `OPTS`
'''
# Take an *args list of modules to import and look for conf.py
# Import conf.py if present
# After gathering all dicts, modify them to merge CLI options
#
# Import python libs
import importlib
import copy


def _ex_final(confs, final, override, key_to_ref, ops_to_ref, globe=False):
    '''
    Scan the configuration datasets, create the final config
    value, and detect collisions
    '''
    for arg in confs:
        for key in confs[arg]:
            ref = f'global.{key}' if globe else f'{arg}.{key}'
            if ref in override:
                s_key = override[ref]['key']
                s_opts = override[ref]['options']
            else:
                s_key = key
                s_opts = confs[arg][key].get('options', [])
            s_opts.append(f'--{s_key}')
            final[s_key] = confs[arg][key]
            if s_opts:
                final[s_key]['options'] = s_opts
            if s_key in key_to_ref:
                key_to_ref[s_key].append(ref)
            else:
                key_to_ref[s_key] = [ref]
            for opt in s_opts:
                if opt in ops_to_ref:
                    ops_to_ref[opt].append(ref)
                else:
                    ops_to_ref = [ref]


def load(
        hub,
        imports,
        override=None,
        cli=None,
        roots=False,
        home_root=None,
        loader='json',
        logs=True):
    '''
    This function takes a list of python packages to load and look for
    respective configs. The configs are then loaded in a non-collision
    way modifying the cli options dynamically.
    The args look for the named <package>.conf python module and then
    looks for dictionaries named after the following convention:

    override = {'<package>.key': 'key': 'new_key', 'options': ['--option1', '--option2']}

    CONFIG: The main configuration for this package - loads to hub.OPT['<import>']
    GLOBAL: Global configs to be used by other packages - loads to hub.OPT['global]
    CLI_CONFIG: Loaded only if this is the only import or if specified in the cli option
    SUBS: Used to define the subcommands, only loaded if this is the cli config
    '''
    if override is None:
        override = {}
    if isinstance(imports, str):
        if cli is None:
            cli = imports
        imports = [imports]
    if home_root is None:
        home_root = cli
    confs = {}
    globe = {}
    final = {}
    collides = []
    key_to_ref = {}
    ops_to_ref = {}
    subs = {}
    for imp in imports:
        cmod = importlib.import_module(f'{imp}.conf')
        if hasattr(cmod, 'CONFIG'):
            confs[imp] = copy.deepcopy(cmod.CONFIG)
        if cli == imp:
            if hasattr(cmod, 'CLI_CONFIG'):
                confs[imp].update(copy.deepcopy(cmod.CLI_CONFIG))
            if hasattr(cmod, 'SUBS'):
                subs = copy.deepcopy(cmod.SUBS)
        if hasattr(cmod, 'GLOBAL'):
            globe[imp] = copy.deepcopy(cmod.GLOBAL)
    if logs:
        lconf = hub.conf.log.init.conf(cli)
        lconf.update(confs[cli])
        confs[cli] = lconf
    _ex_final(confs, final, override, key_to_ref, ops_to_ref)
    _ex_final(globe, final, override, key_to_ref, ops_to_ref, True)
    for opt in ops_to_ref:
        g_count = 0
        if len(ops_to_ref[opt]) > 1:
            collides.append({opt: ops_to_ref[opt]})
    for key in key_to_ref:
        col = []
        for ref in key_to_ref[key]:
            if not ref.startswith('global.'):
                col.append(ref)
        if len(col) > 1:
            collides.append({key: key_to_ref[key]})
    if collides:
        raise KeyError(collides)
    opts = hub.conf.reader.read(final, subs, loader=loader)
    if roots:
        hub.conf.dirs.roots(final.get('root_dir', {}).get('default', '/'), opts, home_root)
        hub.conf.dirs.verify(opts)
    f_opts = {}  # I don't want this to be a defaultdict,
    # if someone tries to add a key willy nilly it should fail
    for key in opts:
        if key == '_subparser_':
            f_opts['_subparser_'] = opts['_subparser_']
        for ref in key_to_ref[key]:
            imp = ref[:ref.rindex('.')]
            if imp not in f_opts:
                f_opts[imp] = {}
            f_opts[imp][key] = opts[key]
    hub.OPT = f_opts
    if logs:
        log_plugin = hub.OPT[cli].get('log_plugin')
        getattr(hub, f'conf.log.{log_plugin}.setup')(hub.OPT[cli])


'''
Used to resolve resolutions to paths on the hub
'''


def last(hub, ref):
    '''
    Takes a string that references the desired ref and returns the last object
    called out in that ref
    '''
    return hub.pop.ref.path(ref)[-1]


def path(hub, ref):
    '''
    Retuns a list of references up to the named ref
    '''
    ret = [hub]
    if isinstance(ref, str):
        ref = ref.split('.')
    for chunk in ref:
        ret.append(getattr(ret[-1], chunk))
    return ret


def create(hub, ref, obj):
    '''
    Create an attribute at a given target using just a ref string and the
    object to be saved at said location. The desired location must already
    exist!

    :param ref: The dot delimited string referencing the target location to
        create the given object on the hub
    :param obj: The object to store at the given reference point
    '''
    if '.' not in ref:
        setattr(hub, ref, obj)
        return
    comps = ref.split('.')
    sub_ref = ref[:ref.rindex('.')]
    var = comps[-1]
    top = hub.pop.ref.last(sub_ref)
    setattr(top, var, obj)



'''
The os module is used to gather configuration options from the OS facility
to send configuration options into applications. In the case of Unix like
systems this translates to the environment variables. On Windows systems
this translates to the registry.
'''
# Import python libs
import os

__virtualname__ = 'os'


def __virtual__(hub):
    '''
    Don't load on Windows, this is for *nix style platforms
    '''
    # TODO: detect if windows
    return True


def gather(hub, defaults):
    '''
    Iterate over the default config data and look for os: True/str options. When set
    gather the option from environment variables is present
    '''
    ret = {}
    for key in defaults:
        if not 'os' in defaults[key]:
            continue
        os_var = defaults[key]['os']
        if os_var is True:
            os_var = key
        os_var = os_var.upper()
        if os_var in os.environ:
            ret[key] = os.environ[os_var]
    return ret



# Import python libs
import asyncio

__virtualname__ = 'cmd'


async def stdout(hub, cmd):
    '''
    Run the passed in function and return the standard out
    '''
    proc = await asyncio.create_subprocess_shell(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            )
    while True:
        line = await proc.stdout.readline()
        if not line:
            break
        yield line
    await proc.wait()



# -*- coding: utf-8 -*-
'''
Provides tools to help unit test projects using pop.
For now, provides mock Hub instances.
'''
import inspect
import copy
from functools import partial
from asynctest.mock import create_autospec

from pop.contract import Contracted
from pop.loader import LoadedMod
from pop.hub import Hub, Sub


class _LookUpTable:
    def __init__(self, *args, **kwargs):
        self._lut = {}
        super().__init__(*args, **kwargs)

    def contains(self, key):
        return self.is_hashable(key) and key in self._lut

    def update(self, key, value):
        if self.is_hashable(key):
            self._lut[key] = value

    def lookup(self, key):
        return self._lut[key]

    def is_hashable(self, key):
        try:
            _ = {key: None}
            return True
        except TypeError:
            return False

    def __len__(self):
        return len(self._lut)


class _LazyPop:
    __lazy_classes = [Hub, Sub, LoadedMod]

    class __Lazy:
        pass

    def __init__(self, obj, lut=None):
        if isinstance(obj, Hub):
            lut = _LookUpTable()
            lut.update('hub', self)
            lut.update(obj, self)
        elif isinstance(obj, Sub):
            obj._load_all()

        self.__lut = lut
        self.__obj = obj
        for attr_name in self.__attr_names():
            setattr(self, attr_name, _LazyPop.__Lazy)

    def __attr_names(self):
        # TODO: '_' - is this actually right? what should I really expose?
        attrs = [attr for attr in self.__obj.__dict__ if not attr.startswith('_')]

        if isinstance(self.__obj, Hub):
            attrs += list(self.__obj._subs)
        elif isinstance(self.__obj, Sub):
            attrs += list(self.__obj._loaded)
            attrs += list(self.__obj._subs)
        elif isinstance(self.__obj, LoadedMod):
            attrs += list(self.__obj._attrs)
        else:
            raise Exception('Standard objects should not be lazy: {}'.format(str(self.__obj)))

        return attrs

    def __getattribute__(self, item):
        if not item.strip('_'):
            raise NotImplementedError
        if '.' in item:
            result = self
            for part in item.split('.').copy():
                result = getattr(result, part)
            return result

        attr = super().__getattribute__(item)

        if attr is _LazyPop.__Lazy:
            orig = getattr(self.__obj, item)

            if self.__lut.contains(orig):
                attr = self.__lut.lookup(orig)
            elif [True for cls in self.__lazy_classes if isinstance(orig, cls)]:
                attr = self.__class__(orig, self.__lut)
            elif isinstance(orig, Contracted):
                attr = self._mock_function(orig)
            else:
                attr = self._mock_attr(orig)

            self.__lut.update(orig, attr)
            setattr(self, item, attr)

        return attr

    def _mock_attr(self, a):
        return create_autospec(a, spec_set=True)

    def _mock_function(self, f):
        raise NotImplementedError()


def strip_hub(f):
    '''
    returns a no-op function with the same function signature... minus the first parameter (hub).
    '''
    if inspect.iscoroutinefunction(f):
        newf = 'async '
    else:
        newf = ''
    newf += 'def {}('.format(f.__name__)
    params = inspect.signature(f).parameters
    new_params = []
    for param in params:
        if params[param].kind is inspect.Parameter.VAR_POSITIONAL:
            new_params.append('*{}'.format(param))
        elif params[param].kind is inspect.Parameter.VAR_KEYWORD:
            new_params.append('**{}'.format(param))
        else:
            new_params.append(param)
        if params[param].default is not inspect.Parameter.empty:
            new_params[-1] += '="has default"'
    newf += ', '.join(new_params[1:])  # skip hub
    newf += '): pass'

    scope = {}
    exec(newf, scope)

    return scope[f.__name__]


class MockHub(_LazyPop):
    '''
    Provides mocks mirroring a real hub::

        hub.sub.mod.fn()  # mock
        hub.sub.mod.attr  # mock
    '''
    def _mock_function(self, f):
        return create_autospec(strip_hub(f.func), spec_set=True)


class NoContractHub(_LazyPop):
    '''
    Provides access to real functions, bypassing contracts and mocking attributes::

        hub.sub.mod.fn()  # executes real function, no contracts
        hub.sub.mod.attr  # mock
    '''
    def _mock_function(self, f):
        return partial(f.func, self._LazyPop__lut.lookup('hub'))


class MockContracted:
    '''
    Creates a new contracted, but using a mock function.
    The mock function is masked to look just like the real function.

    Look up/set attributes first on the new contracted, then pass through to the mock.
    '''

    def __init__(self, c):
        mock_func = create_autospec(c.func, spec_set=True)
        mock_func.__module__ = c.func.__module__
        mock_func.__dict__.update(copy.deepcopy(c.func.__dict__))
        self.__dict__['contracted'] = Contracted(c.hub, c.contracts, mock_func, c.ref, c.name)
        self.signature = c.signature

    def __call__(self, *args, **kwargs):
        return self.contracted(*args, **kwargs)

    def __getattr__(self, attr):
        if attr in self.contracted.__dict__:
            # allow access to contracted variables
            return getattr(self.contracted, attr)
        else:
            # but pass through to mock functions otherwise
            return getattr(self.contracted.func, attr)

    def __setattr__(self, name, value):
        if name in self.contracted.__dict__:
            # allow access to contracted variables
            setattr(self.contracted, name, value)
        else:
            # but pass through to mock functions otherwise
            setattr(self.contracted.func, name, value)


class ContractHub(_LazyPop):
    '''
    Runs a call through the contract system, but the function is a mock. Mostly useful for integration tests:

        hub.sub.mod.fn()  # executes mock function, real contracts
        hub.sub.mod.attr  # mock

    You can verify what parameters are passed to a function after going through loaded contracts::

        contract_hub.sub.mod.fn('foo')
        assert contract_hub.sub.mod.fn.called_with('bar')

    --------------------------------

    You can view or modify the contracts that will be executed on one function for a test - but first:
    MODIFYING CONTRACTS THIS WAY IS NOT SAFE ON REAL HUBS AND OTHER TESTING HUB VARIANTS!

    I have previously thought of modifying contracts with mocks, only to realize what I really want is to
    unit test a specific contract. Think twice before using this functionality.

    --------------------------------

    The contract modules are visible via hub.sub.mod.fn.contracts, and the contract functions that will
    be called, wrapping fn are visible via hub.sub.mod.fn.contract_functions. It is safe to modify the
    contracts list or contract_functions dict only on a ContractHub.

    Examine that the first contract function to be called is 'foo.pre_fn', then bypass it::

        assert contract_hub.sub.mod.fn.contract_functions['pre'][0].__module__ is 'foo'
        assert contract_hub.sub.mod.fn.contract_functions['pre'][0].__name__ is 'pre_fn'
        hub.sub.mod.fn.contract_functions['pre'][0] = create_autospec(hub.sub.mod.fn.contract_functions['pre'][0])

    Assert that one contract will be called before another::

        assert contract_hub.sub.mod.fn.contracts.index(contract1) < contract_hub.sub.mod.fn.contracts.index(contract2)
    '''
    def _mock_function(self, f):
        return MockContracted(f)



# -*- coding: utf-8 -*-
'''
The pop top level module
'''

# Import python libs
import os

INSTALL_DIR = os.path.dirname(os.path.realpath(__file__))


# ----- Detect System Encoding -------------------------------------------------------------------------------------->
# This will install a globally available variable which will hold the detected encoding
def __define_global_system_encoding_variable__():
    import sys
    # sys.stdin.encoding is the most trustworthy source of the system encoding, though, if
    # pop is being imported after being daemonized, this information is lost and reset to None
    encoding = None

    if not sys.platform.startswith('win') and sys.stdin is not None:
        # On linux we can rely on sys.stdin for the encoding since it
        # most commonly matches the filesystem encoding. This however
        # does not apply to windows
        encoding = sys.stdin.encoding

    if not encoding:
        # If the system is properly configured this should return a valid
        # encoding. MS Windows has problems with this and reports the wrong
        # encoding
        import locale
        try:
            encoding = locale.getdefaultlocale()[-1]
        except ValueError:
            # A bad locale setting was most likely found:
            #   https://github.com/saltstack/salt/issues/26063
            pass

        # This is now garbage collectible
        del locale

        if not encoding:
            if sys.platform.startswith('darwin'):
                # Mac OS X uses UTF-8
                encoding = 'utf-8'
            elif sys.platform.startswith('win'):
                # Windows uses a configurable encoding; on Windows, Python uses the name “mbcs”
                # to refer to whatever the currently configured encoding is.
                encoding = 'mbcs'
            else:
                # This is most likely ascii which is not the best but we were
                # unable to find a better encoding. If this fails, we fall all
                # the way back to ascii
                # On linux default to ascii as a last resort
                encoding = sys.getdefaultencoding() or 'ascii'

    try:
        # Return the detected encoding
        return encoding
    finally:
        # This is now garbage collectible
        del sys
        del encoding


PACK_SYSTEM_ENCODING = __define_global_system_encoding_variable__()

# This is now garbage collectible
del __define_global_system_encoding_variable__
# <---- Detect System Encoding ---------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
'''
Load the files detected from the scanner
'''
# Import Python libs
import os
import sys
import inspect
import importlib
import importlib.util
import importlib.machinery
import traceback as stdlib_traceback

# Import pop libs
import pop.exc
import pop.contract


class LoadError:
    '''
    Errors from the loader are contained herein
    '''
    __slots__ = ('edict', 'traceback')

    def __init__(self, msg, exception=None, traceback=None, verror=None):
        self.edict = {
                'msg': msg,
                'exception': exception,
                'verror': verror,
                }
        self.traceback = traceback

    def __call__(self):
        '''
        Return the error cases
        '''
        return self.edict

    def __getattr__(self, attr):
        '''
        Return a lambda that returns the edict
        '''
        return self.__calling_load_error__

    def __calling_load_error__(self, *args, **kwargs):  # pylint: disable=unused-argument
        if self.edict['verror']:
            error = '{0[msg]}: {0[verror]}'.format(self())
            raise pop.exc.PopError(error)
        error = '{0[msg]}: {0[exception]!r}'.format(self())
        if self.traceback:
            error += '\n' + self.traceback
        raise pop.exc.PopError(error)

    def __repr__(self):
        return '<{} edict={!r}>'.format(self.__class__.__name__, self.edict)


def load_mod(modname, form, path):
    '''
    Load a single module
    '''
    this = sys.modules[__name__]
    return getattr(this, form)(modname, path)


def _generate_module(name):
    '''
    Generate a module at runtime and insert it in sys.modules
    '''
    if name in sys.modules:
        return sys.modules[name]

    code = "'''POP sub auto generated parent module for {0}'''".format(name.split('.')[-1])
    module = stdlib_imp.new_module(name)
    exec(code, module.__dict__)  # pylint: disable=exec-used
    sys.modules[name] = module
    return module


def _populate_sys_modules(mod):
    '''
    This is a hack to populate sys.modules with the modules that pop loads
    while making sure that parent modules have the attribute for the child
    modules.
    '''
    mod_parts = mod.split('.')
    imp_mod = mod_parts.pop(0)
    gen_mod = _generate_module(imp_mod)
    while True:
        if not mod_parts:
            break
        part = mod_parts.pop(0)
        imp_mod += '.' + part
        gen_child_mod = _generate_module(imp_mod)
        setattr(gen_mod, part, gen_child_mod)
        gen_mod = gen_child_mod


def ext(modname, path):
    '''
    Attempt to load the named python modules
    '''
    modname = '.'.join(modname.split('.')[:-1])
    if modname in sys.modules:
        return sys.modules[modname]
    _populate_sys_modules(modname)
    try:
        efl = importlib.machinery.ExtensionFileLoader(modname, path)
        mod = efl.create_module(importlib.util.find_spec(modname))
        return efl.exec_module(mod)
    except Exception as exc:  # pylint: disable=broad-except
        return LoadError(
                'Failed to load python module {} at path {}'.format(modname, path),
                exception=exc,
                traceback=stdlib_traceback.format_exc())


def python(modname, path):
    '''
    Attempt to load the named python modules
    '''
    if modname in sys.modules:
        return sys.modules[modname]
    try:
        sfl = importlib.machinery.SourceFileLoader(modname, path)
        return sfl.load_module()
    except Exception as exc:  # pylint: disable=broad-except
        return LoadError(
                'Failed to load python module {} at path {}'.format(modname, path),
                exception=exc,
                traceback=stdlib_traceback.format_exc())


def load_virtual(hub, virtual, mod, bname):
    '''
    Run the virtual function to name the module and check for all loader
    errors
    '''
    base_name = os.path.basename(bname)
    if '.' in base_name:
        base_name = base_name.split('.')[0]
    try:
        name = mod.__virtualname__
    except Exception:  # pylint: disable=broad-except
        name = base_name
    if isinstance(mod, LoadError):
        # The mod is a LoadError instance.
        # Return the load error with name as the base_name because another
        # module is still allowed to load under the same __virtualname__
        # but also return the vname information
        return {'name': base_name, 'vname': name, 'error': mod}

    if not virtual:
        # __virtual__ is not to be processed. Return now!
        return {'name': base_name}

    if not hasattr(mod, '__virtual__'):
        # No __virtual__ processing is required.
        # Return the mod's name as the defined __virtualname__ if defined,
        # else, the base_name
        return {'name': name}

    try:
        vret = mod.__virtual__(hub)
    except Exception as exc:  # pylint: disable=broad-except
        err = LoadError(
                'Virtual threw exception in mod {}'.format(bname),
                exception=exc,
                traceback=stdlib_traceback.format_exc())
        # Return the load error with name as the base_name because another
        # module is still allowed to load under the same __virtualname__
        # but also return the vname information
        return {'name': base_name, 'vname': name, 'error': err}

    if vret is True:
        # No problems occurred, module is allowed to load
        # Return the mod's name as the defined __virtualname__ if defined,
        # else, the base_name
        return {'name': name}

    if vret is False:
        # __virtual__ explicitly disabled the loading of this module
        err = LoadError(
                'Module {} returned virtual FALSE'.format(bname),
                verror=vret)
        # Return the load error with name as the base_name because another
        # module is still allowed to load under the same __virtualname__
        # but also return the vname information
        return {'name': base_name, 'vname': name, 'error': err}

    # Anything else besides True/False should be considered a LoadError
    err = LoadError(
            'Module {} returned virtual error'.format(bname),
            verror=vret)
    # Return the load error with name as the base_name because another
    # module is still allowed to load under the same __virtualname__
    # but also return the vname information
    return {'name': base_name, 'vname': name, 'error': err}


def mod_init(hub, mod):
    '''
    Process module's __init__ function if defined
    '''
    if '__init__' in dir(mod):
        mod.__init__(hub)


def prep_loaded_mod(this_sub, mod, mod_name, contracts):
    '''
    Read the attributes of a python module and create a LoadedMod, which resolves
    aliases and omits objects that should not be exposed.
    '''
    # pylint: disable=protected-access
    lmod = LoadedMod(mod_name)
    ref = f'{this_sub._subname}.{mod_name}'  # getattr(hub, ref) should resolve to this module
    for attr in getattr(mod, '__load__', dir(mod)):
        name = getattr(mod, '__func_alias__', {}).get(attr, attr)
        func = getattr(mod, attr)
        if not this_sub._omit_vars:
            if not inspect.isfunction(func) and not inspect.isclass(func) and \
                    type(func).__name__ != 'cython_function_or_method':
                lmod._vars[name] = func
                lmod._attrs[name] = func
                continue
        if attr.startswith(this_sub._omit_start):
            continue
        if attr.endswith(this_sub._omit_end):
            continue
        if inspect.isfunction(func) or inspect.isbuiltin(func) or \
                type(func).__name__ == 'cython_function_or_method':
            obj = pop.contract.Contracted(this_sub._hub, contracts, func, ref, name)
            if not this_sub._omit_func:
                if this_sub._pypath and not func.__module__.startswith(mod.__name__):
                    # We're only interested in functions defined in this module, not
                    # imported functions
                    continue
                lmod._funcs[name] = obj
                lmod._attrs[name] = obj
        else:
            klass = func
            if not this_sub._omit_class and inspect.isclass(klass):
                # We're only interested in classes defined in this module, not
                # imported classes
                if not klass.__module__.startswith(mod.__name__):
                    continue
                lmod._classes[name] = klass
                lmod._attrs[name] = klass
    return lmod


class LoadedMod:
    '''
    The LoadedMod class allows for the module loaded onto the sub to return
    custom sequencing, for instance it can be iterated over to return all
    functions
    '''
    def __init__(self, name):
        self.__sub_name__ = name
        self._vars = {}
        self._funcs = {}
        self._classes = {}
        self._attrs = {}

    def __getattr__(self, item):
        if item in self._attrs:
            return self._attrs[item]
        raise AttributeError(item)

    def __iter__(self):
        keys = sorted(self._funcs)
        ret = []
        for key in keys:
            ret.append(self._funcs[key])
        return iter(ret)

    def __dir__(self):
        # TODO: This should return finite set attrs as well as dunder attrs
        ret = list(self._attrs.keys())
        ret.extend(['__name__', '_vars', '_funcs', '_classes', '_attrs'])
        return ret



'''
Execute functions or load subs on the workers in the named worker pool
'''
# import python libs
import asyncio
import os
# Import third party libs
import msgpack


async def add_sub(hub, worker_name, *args, **kwargs):
    '''
    Tell all of the worker in the named pool to load the given sub,

    This function takes all of the same arguments as hub.pop.sub.add
    '''
    ret = {}
    workers = hub.proc.Workers[worker_name]
    for ind in workers:
        payload = {'fun': 'sub', 'args': args, 'kwargs': kwargs}
        # TODO: Make these futures to the run at the same time
        async for chunk in hub.proc.run.send(workers[ind], payload):
            ret[ind] = chunk
    hub.proc.WorkersTrack[worker_name]['subs'].append({'args': args, 'kwargs': kwargs})
    return ret


async def add_proc(hub, worker_name):
    '''
    Add a single process to the worker pool, also make sure that
    '''
    # grab and extrapolate the data we need
    ret_ref = hub.proc.WorkersTrack[worker_name]['ret_ref']
    sock_dir = hub.proc.WorkersTrack[worker_name]['sock_dir']
    workers = hub.proc.Workers[worker_name]
    ind = len(workers) + 1
    for s_ind in range(len(workers) + 1):
        if s_ind not in workers:
            ind = s_ind
    hub.proc.init.mk_proc(ind, workers, ret_ref, sock_dir)
    # Make sure the process is up with a live socket
    while True:
        if os.path.exists(workers[ind]['path']):
            break
        await asyncio.sleep(0.01)
    # Add all of the subs that have been added to processes in this pool
    for sub in hub.proc.WorkersTrack[worker_name]['subs']:
        payload = {'fun': 'sub', 'args': sub['args'], 'kwargs': sub['kwargs']}
        async for chunk in hub.proc.run.send(workers[ind], payload):
            pass
    return ind


async def pub(hub, worker_name, func_ref, *args, **kwargs):
    '''
    Execute the given function reference on ALL the workers in the given
    worker pool and return the return data from each.

    Pass in the arguments for the function, keep in mind that the sub needs
    to be loaded into the workers for a function to be available via
    hub.proc.run.add_sub
    '''
    workers = hub.proc.Workers[worker_name]
    ret = {}
    for ind in workers:
        payload = {'fun': 'run', 'ref': func_ref, 'args': args, 'kwargs': kwargs}
        # TODO: Make these futures to the run at the same time
        async for chunk in hub.proc.run.send(workers[ind], payload):
            ret[ind] = chunk
    return ret


async def set_attr(hub, worker_name, ref, value):
    '''
    Set the given attribute to the given location on the hub of all
    worker procs
    '''
    workers = hub.proc.Workers[worker_name]
    ret = {}
    for ind in workers:
        payload = {'fun': 'setattr', 'ref': ref, 'value': value}
        # TODO: Make these futures to the run at the same time
        async for chunk in hub.proc.run.send(workers[ind], payload):
            ret[ind] = chunk
    return ret


async def ind_func(hub, worker_name, _ind, func_ref, *args, **kwargs):
    '''
    Execute the function on the indexed process within the named worker pool
    '''
    workers = hub.proc.Workers[worker_name]
    worker = workers[_ind]
    payload = {'fun': 'run', 'ref': func_ref, 'args': args, 'kwargs': kwargs}
    async for ret in hub.proc.run.send(worker, payload):
        return ret


async def func(hub, worker_name, func_ref, *args, **kwargs):
    '''
    Execute the given function reference on one worker in the given worker
    pool and return the return data.

    Pass in the arguments for the function, keep in mind that the sub needs
    to be loaded into the workers for a function to be available via
    hub.proc.run.add_sub
    '''
    ind, coro = await hub.proc.run.track_func(worker_name, func_ref, *args, **kwargs)
    return await coro


async def track_func(hub, worker_name, func_ref, *args, **kwargs):
    '''
    Run a function and return the index of the worker that the function was
    executed on and a coroutine to track
    '''
    w_iter = hub.proc.WorkersIter[worker_name]
    ind = next(w_iter)
    coro = hub.proc.run.ind_func(worker_name, ind, func_ref, *args, **kwargs)
    return ind, coro


async def gen(hub, worker_name, func_ref, *args, **kwargs):
    '''
    Execute a generator function reference within one worker within the given
    worker pool.

    Like `func` the sub needs to be made available to all workers first
    '''
    ind, coro = await hub.proc.run.track_gen(worker_name, func_ref, *args, **kwargs)
    async for chunk in coro:
        yield chunk


async def track_gen(hub, worker_name, func_ref, *args, **kwargs):
    '''
    Return an iterable coroutine and the index executed on
    '''
    w_iter = hub.proc.WorkersIter[worker_name]
    ind = next(w_iter)
    coro = hub.proc.run.ind_gen(worker_name, ind, func_ref, *args, **kwargs)
    return ind, coro


async def ind_gen(hub, worker_name, _ind, func_ref, *args, **kwargs):
    '''
    run the given iterator on the defined index
    '''
    workers = hub.proc.Workers[worker_name]
    worker = workers[_ind]
    payload = {'fun': 'gen', 'ref': func_ref, 'args': args, 'kwargs': kwargs}
    async for chunk in hub.proc.run.send(worker, payload):
        yield chunk


async def send(hub, worker, payload):
    '''
    Send the given payload to the given worker, yield iterations based on the
    returns from the remote.
    '''
    mp = msgpack.dumps(payload, use_bin_type=True)
    mp += hub.proc.DELIM
    reader, writer = await asyncio.open_unix_connection(path=worker['path'])
    writer.write(mp)
    await writer.drain()
    final_ret = True
    while True:
        ret = await reader.readuntil(hub.proc.DELIM)
        p_ret = ret[:-len(hub.proc.DELIM)]
        i_flag = p_ret[-1:]
        ret = msgpack.loads(p_ret[:-1], raw=False)
        if i_flag == hub.proc.D_FLAG:
            # break for the end of the sequence
            break
        yield ret
        final_ret = False
    if final_ret:
        yield ret



# -*- coding: utf-8 -*-
'''
Tools to work with dicts
'''

__virtualname__ = 'dicts'


def traverse(hub, data, key, default=None, delimiter=':'):
    '''
    traverse a dict or list using a colon-delimited (or otherwise delimited,
    using the 'delimiter' param) target string. the target 'foo:bar:0' will
    return data['foo']['bar'][0] if this value exists, and will otherwise
    return the dict in the default argument.
    function will automatically determine the target type.
    the target 'foo:bar:0' will return data['foo']['bar'][0] if data like
    {'foo':{'bar':['baz']}} , if data like {'foo':{'bar':{'0':'baz'}}}
    then return data['foo']['bar']['0']
    '''
    for each in key.split(delimiter):
        if isinstance(data, list):
            try:
                idx = int(each)
            except ValueError:
                embed_match = False
                # Index was not numeric, lets look at any embedded dicts
                for embedded in (x for x in data if isinstance(x, dict)):
                    try:
                        data = embedded[each]
                        embed_match = True
                        break
                    except KeyError:
                        pass
                if not embed_match:
                    # No embedded dicts matched, return the default
                    return default
            else:
                try:
                    data = data[idx]
                except IndexError:
                    return default
        else:
            try:
                data = data[each]
            except (KeyError, TypeError):
                return default
    return data



# -*- coding: utf-8 -*-
'''
Find directories
'''
# Import python libs
import os
import sys
import importlib


def dir_list(subname, p_name, pypath=None, static=None):
    '''
    Return the directories to look for modules in, pypath specifies files
    relative to an installed python package, static is for static dirs
    '''
    ret = []
    for path in pypath:
        mod = importlib.import_module(path)
        for m_path in mod.__path__:
            # If we are inside of an executable the path will be different
            ret.append(m_path)
    ret.extend(static)
    return ret


def inline_dirs(dirs, subdir):
    '''
    Look for the named subdir in the list of dirs
    '''
    ret = []
    for dir_ in dirs:
        check = os.path.join(dir_, subdir)
        if os.path.isdir(check):
            ret.append(check)
    return ret


def dynamic_dirs():
    '''
    Iterate over the available python package imports and look for configured
    dynamic dirs
    '''
    dirs = []
    ret = {}
    for dir_ in sys.path:
        if not os.path.isdir(dir_):
            continue
        for sub in os.listdir(dir_):
            full = os.path.join(dir_, sub)
            if full.endswith('.egg-link'):
                with open(full) as rfh:
                    dirs.append(rfh.read().strip())
            if os.path.isdir(full):
                dirs.append(full)
    for dir_ in dirs:
        conf = os.path.join(dir_, 'conf.py')
        if not os.path.isfile(conf):
            continue
        try:
            modname =f'{os.path.basename(dir_)}.conf'
            sfl = importlib.machinery.SourceFileLoader(modname, conf)
            cmod = sfl.load_module()
        except Exception:
            continue
        if hasattr(cmod, 'DYNE'):
            if not isinstance(cmod.DYNE, dict):
                continue
            for name, paths in cmod.DYNE.items():
                if not isinstance(paths, list):
                    continue
                if name not in ret:
                    ret[name] = []
                for path in paths:
                    ret[name].append(os.path.join(dir_, path.replace('.', os.sep)))
    return ret



# -*- coding: utf-8 -*-
'''
Translate an options data structure into command line args
'''

# Import python libs
import sys
import inspect
import argparse
import functools
import collections

__virtualname__ = 'args'
__contracts__ = [__virtualname__]


class ActionWrapper:
    '''
    This class wraps argparse.Action instances in order to mark arguments passed
    on CLI as explicitly passed
    '''
    def __init__(self, action):
        self._action = action
        functools.update_wrapper(self, action)

    def __call__(self, parser, namespace, values, option_string):
        # Let's store the call to this option as an explicit CLI call for later
        # use when overwriting any configuration settings on file with those
        # from CLI
        if getattr(parser, '_explicit_cli_args_', None) is None:
            setattr(parser, '_explicit_cli_args_', set())
        parser._explicit_cli_args_.add(self._action.dest)  # pylint: disable=protected-access
        # Carry on regular operation
        return self._action(parser, namespace, values, option_string)

    def __getattribute__(self, name):
        if name == '_action':
            return object.__getattribute__(self, name)
        # Proxy any attribute's search to the _action instance
        return getattr(self._action, name)

    def __repr__(self):
        return repr(self._action)


class ActionClassWrapper:
    '''
    This class wraps argparse.Action classes in order to mark arguments passed
    on CLI as explicitly passed
    '''
    def __init__(self, klass):
        self._klass = klass

    def __call__(self, *args, **kwargs):
        return ActionWrapper(self._klass(*args, **kwargs))

    def __repr__(self):
        return repr(self._klass)

    def __getattribute__(self, name):
        if name == '_klass':
            return object.__getattribute__(self, name)
        # Proxy any attributes search to the _klass instance
        return getattr(self._klass, name)


class ArgumentParser(argparse.ArgumentParser):
    def register(self, name, value, obj):  # pylint: disable=arguments-differ
        if name == 'action':
            # Let's wrap it on our action class wrapper so we can latter store
            # which options were explicitly passed from CLI
            return super(ArgumentParser, self).register(
                name,
                value,
                ActionClassWrapper(obj)
            )
        return super(ArgumentParser, self).register(name, value, obj)

    def parse_known_args(self, args=None, namespace=None):
        namespace, arg_strings = super().parse_known_args(args, namespace)
        explicit_cli_args = getattr(self, '_explicit_cli_args_', set())
        if '_explicit_cli_args_' not in namespace:
            setattr(namespace, '_explicit_cli_args_', set())
        namespace._explicit_cli_args_.update(explicit_cli_args)
        return namespace, arg_strings


def __init__(hub):
    '''
    Set up the local memory copy of the parser
    '''
    hub.conf._mem['args'] = {}


def _init_parser(hub, opts):
    if 'parser' not in hub.conf._mem['args']:
        # Instantiate the parser
        hub.conf._mem['args']['parser'] = ArgumentParser(**opts.get('_argparser_', {}))


def _keys(opts):
    '''
    Return the keys in the right order
    '''
    if isinstance(opts, collections.OrderedDict):
        return sorted(list(opts),
                      key=lambda k: opts[k].get('display_priority', sys.maxsize))
    return sorted(opts,
                  key=lambda k: (opts[k].get('display_priority', sys.maxsize), k))


def subs(hub, opts):
    '''
    Set up sub parsers, if using sub parsers this needs to be called
    before calling setup.

    opts dict:
        <sub_title>:
            [desc]: 'Some subparser'
            help: 'subparser!'
    '''
    _init_parser(hub, opts)
    hub.conf._mem['args']['sub'] = hub.conf._mem['args']['parser'].add_subparsers(dest='_subparser_')
    hub.conf._mem['args']['subs'] = {}
    for arg in _keys(opts):
        if arg in ('_argparser_',):
            continue
        comps = opts[arg]
        kwargs = {}
        if 'help' in comps:
            kwargs['help'] = comps['help']
        if 'desc' in comps:
            kwargs['description'] = comps['desc']
        hub.conf._mem['args']['subs'][arg] = hub.conf._mem['args']['sub'].add_parser(arg, **kwargs)
    return {'result': True, 'return': True}


def setup(hub, opts):
    '''
    Take in a pre-defined opts dict and translate it to args

    opts dict:
        <arg>:
            [group]: foo
            [default]: bar
            [action]: store_true
            [options]: # arg will be turned into --arg
              - '-A'
              - '--cheese'
            [choices]:
              - foo
              - bar
              - baz
            [nargs]: +
            [type]: int
            [dest]: cheese
            help: Some great help message
    '''
    _init_parser(hub, opts)
    defaults = {}
    groups = {}
    ex_groups = {}
    for arg in _keys(opts):
        if arg in ('_argparser_',):
            continue
        comps = opts[arg]
        positional = comps.pop('positional', False)
        if positional:
            args = [arg]
        else:
            long_opts = ['--{}'.format(arg.replace('_', '-'))]
            short_opts = []
            for o_str in comps.get('options', []):
                if not o_str.startswith('--') and o_str.startswith('-'):
                    short_opts.append(o_str)
                    continue
                long_opts.append(o_str)
            args = short_opts + long_opts
        kwargs = {}
        kwargs['action'] = action = comps.get('action', None)

        if action is None:
            # Non existing option defaults to a StoreAction in argparse
            action = hub.conf._mem['args']['parser']._registry_get('action', action)  # pylint: disable=protected-access

        if isinstance(action, str):
            signature = inspect.signature(
                hub.conf._mem['args']['parser']._registry_get('action', action).__init__)  # pylint: disable=protected-access
        else:
            signature = inspect.signature(action.__init__)

        for param in signature.parameters:
            if param == 'self' or param not in comps:
                continue
            if param == 'dest':
                kwargs['dest'] = comps.get('dest', arg)
                continue
            if param == 'help':
                kwargs['help'] = comps.get('help', 'THIS NEEDS SOME DOCUMENTATION!!')
                continue
            if param == 'default':
                defaults[comps.get('dest', arg)] = comps[param]
            kwargs[param] = comps[param]

        if 'group' in comps:
            group = comps['group']
            if group not in groups:
                groups[group] = hub.conf._mem['args']['parser'].add_argument_group(group)
            groups[group].add_argument(*args, **kwargs)
            continue
        if 'ex_group' in comps:
            group = comps['ex_group']
            if group not in ex_groups:
                ex_groups[group] = hub.conf._mem['args']['parser'].add_mutually_exclusive_group()
            ex_groups[group].add_argument(*args, **kwargs)
            continue
        if 'sub' in comps:
            sub = comps['sub']
            sparse = hub.conf._mem['args']['subs'].get(sub)
            if not sparse:
                # Maybe raise exception here? Malformed config?
                continue
            sparse.add_argument(*args, **kwargs)
            continue
        hub.conf._mem['args']['parser'].add_argument(*args, **kwargs)
    return {'result': True, 'return': defaults}


def parse(hub, args=None, namespace=None, only_parse_known_arguments=False):
    '''
    Parse the command line options
    '''
    if only_parse_known_arguments:
        opts, unknown_args = hub.conf._mem['args']['parser'].parse_known_args(args, namespace)
        opts_dict = opts.__dict__
        opts_dict['_unknown_args_'] = unknown_args
    else:
        opts = hub.conf._mem['args']['parser'].parse_args(args, namespace)
        opts_dict = opts.__dict__
    return {'result': True, 'return': opts_dict}


def render(hub, defaults, cli_opts, explicit_cli_args):
    '''
    For options specified as such, take the string passed into the cli and
    render it using the specified render flag
    '''
    for key in explicit_cli_args:
        rend = defaults.get(key, {}).get('render')
        if rend:
            ref = f'conf.{rend}.render'
            cli_opts[key] = hub.pop.ref.last(ref)(cli_opts[key])
    return cli_opts



# Import python libs
import inspect
# Import pop libs
import pop.exc


def contract(hub, raws, mod):  # pylint: disable=unused-argument
    '''
    Verify module level contract - functions only
    '''
    sig_errs = []
    sig_miss = []
    mname = mod.__name__
    for raw in raws:
        for fun in raw._funcs:
            if fun.startswith('sig_'):
                tfun = fun[4:]
                if tfun not in mod._funcs:
                    sig_miss.append(tfun)
                    continue
                sig_errs.extend(sig(mod._funcs[tfun].func, raw._funcs[fun].func))
    if sig_errs or sig_miss:
        msg = ''
        if sig_errs:
            msg += f'Signature Errors in {mname}:\n'
            for err in sig_errs:
                msg += f'{err}\n'
        if sig_miss:
            msg += f'Signature Functions Missing in {mname}:\n'
            for err in sig_miss:
                msg += f'{err}\n'
        msg = msg.strip()
        raise pop.exc.ContractSigException(msg)


def sig_map(ver):
    '''
    Generates the map dict for the signature verification
    '''
    vsig = inspect.signature(ver)
    vparams = list(vsig.parameters.values())
    vdat = {'args': [], 'v_pos': -1, 'kw': [], 'kwargs': False, 'ann': {}}
    for ind in range(len(vparams)):
        param = vparams[ind]
        val = param.kind.value
        name = param.name
        if val == 0 or val == 1:
            vdat['args'].append(name)
            if param.default != inspect._empty: # Is a KW, can be inside of **kwargs
                vdat['kw'].append(name)
        elif val == 2:
            vdat['v_pos'] = ind
        elif val == 3:
            vdat['kw'].append(name)
        elif val == 4:
            vdat['kwargs'] = ind
        if param.annotation != inspect._empty:
            vdat['ann'][name] = param.annotation
    return vdat


def sig(func, ver):
    '''
    Takes 2 functions, the first function is verified to have a parameter signature
    compatible with the second function
    '''
    errors = []
    fsig = inspect.signature(func)
    fparams = list(fsig.parameters.values())
    vdat = sig_map(ver)
    arg_len = len(vdat['args'])
    v_pos = False
    for ind in range(len(fparams)):
        param = fparams[ind]
        val = param.kind.value
        name = param.name
        has_default = param.default != inspect._empty
        ann = param.annotation
        vann = vdat['ann'].get(name, inspect._empty)
        if vann != ann:
            errors.append(f'Parameter, "{name}" is type "{str(ann)}" not "{str(vann)}"')
        if val == 2:
            v_pos = True
        if val == 0 or val == 1:
            if ind >= arg_len:  # Past available positional args
                if not vdat['v_pos'] == -1:  # Has a *args
                    if ind >= vdat['v_pos'] and v_pos:
                        # Invalid unless it is a kw
                        if not name in vdat['kw']:
                            # Is a kw
                            errors.append(f'Parameter "{name}" is invalid')
                        if vdat['kwargs'] is False:
                            errors.append(f'Parameter "{name}" not defined as kw only')
                        continue
                elif vdat['kwargs'] is not False and not has_default:
                    errors.append(f'Parameter "{name}" is past available positional params')
                elif vdat['kwargs'] is False:
                    errors.append(f'Parameter "{name}" is past available positional params')
            else:
                v_param = vdat['args'][ind]
                if v_param != name:
                    errors.append(f'Parameter "{name}" does not have the correct name: {v_param}')
        if val == 2:
            if ind < vdat['v_pos']:
                errors.append(f'Parameter "{name}" is not in the correct position for *args')
        if val == 3:
            if name not in vdat['kw'] and not vdat['kwargs']:
                errors.append(f'Parameter "{name}" is not available as a kwarg')
        if val == 4:
            if vdat['kwargs'] is False:
                errors.append(f'Kwargs are not permitted as a parameter')
    return errors



'''
The Proc sub is used to spin up worker processes that run hub referenced
coroutines.
'''
# Import python libs
import os
import sys
import atexit
import itertools
import asyncio
import subprocess
# Import third party libs
import msgpack


def __init__(hub):
    '''
    Create constants used by the client and server side of procs
    '''
    hub.proc.DELIM = b'd\xff\xcfCO)\xfe='
    hub.proc.D_FLAG = b'D'
    hub.proc.I_FLAG = b'I'
    hub.proc.Workers = {}
    hub.proc.WorkersIter = {}
    hub.proc.WorkersTrack = {}


def _get_cmd(hub, ind, ref, ret_ref, sock_dir):
    '''
    Return the shell command to execute that will start up the worker
    '''
    code = 'import sys; '
    code += 'import pop.hub; '
    code += 'hub = pop.hub.Hub(); '
    code += 'hub.pop.sub.add("pop.mods.proc"); '
    code += f'hub.proc.worker.start("{sock_dir}", "{ind}", "{ref}", "{ret_ref}")'
    cmd = f'{sys.executable} -c \'{code}\''
    return cmd


def mk_proc(hub, ind, workers, ret_ref, sock_dir):
    '''
    Create the process and add it to the passed in workers dict at the
    specified index
    '''
    ref = os.urandom(3).hex() + '.sock'
    workers[ind] = {'ref': ref}
    workers[ind]['path'] = os.path.join(sock_dir, ref)
    cmd = _get_cmd(hub, ind, ref, ret_ref, sock_dir)
    workers[ind]['proc'] = subprocess.Popen(cmd, shell=True)
    workers[ind]['pid'] = workers[ind]['proc'].pid


async def pool(hub, num, name='Workers', callback=None, sock_dir=None):
    '''
    Create a new local pool of process based workers

    :param num: The number of processes to add to this pool
    :param ref: The location on the hub to create the Workers dict used to
        store the worker pool, defaults to `hub.pop.proc.Workers`
    :param callback: The pop ref to call when the process communicates
        back
    '''
    ret_ref = os.urandom(3).hex() + '.sock'
    ret_sock_path = os.path.join(sock_dir, ret_ref)
    if not hub.proc.Tracker:
        hub.proc.init.mk_tracker()
    workers = {}
    if callback:
        await asyncio.start_unix_server(
                hub.proc.init.ret_work(callback),
                path=ret_sock_path)
    for ind in range(num):
        hub.proc.init.mk_proc(ind, workers, ret_ref, sock_dir)
    w_iter = itertools.cycle(workers)
    hub.proc.Workers[name] = workers
    hub.proc.WorkersIter[name] = w_iter
    hub.proc.WorkersTrack[name] = {
        'subs': [],
        'ret_ref': ret_ref,
        'sock_dir': sock_dir}
    up = set()
    while True:
        for ind in workers:
            if os.path.exists(workers[ind]['path']):
                up.add(ind)
        if len(up) == num:
            break
        await asyncio.sleep(0.01)
    # TODO: This seems to be spawning extra procs, this should be fixed
    #asyncio.ensure_future(hub.proc.init.maintain(name))


async def maintain(hub, name):
    '''
    Keep an eye on these processes
    '''
    workers = hub.proc.Workers[name]
    while True:
        for ind, data in workers.items():
            if not data['proc'].poll():
                hub.proc.init.mk_proc(ind, workers)
        await asyncio.sleep(2)


def mk_tracker(hub):
    '''
    Create the process tracker, this simply makes a data structure to hold
    process references and sets them to be terminated when the system is
    shutdown.
    '''
    hub.proc.Tracker = True
    atexit.register(hub.proc.init.clean)


def clean(hub):
    '''
    Clean up the processes registered in the tracker
    '''
    for name, workers in hub.proc.Workers.items():
        for ind in workers:
            workers[ind]['proc'].terminate()


def ret_work(hub, callback):
    async def work(reader, writer):
        '''
        Process the incoming work
        '''
        inbound = await reader.readuntil(hub.proc.DELIM)
        inbound = inbound[:-len(hub.proc.DELIM)]
        payload = msgpack.loads(inbound, raw=False)
        ret = await callback(payload)
        ret = msgpack.dumps(ret, use_bin_type=True)
        ret += hub.proc.DELIM
        writer.write(ret)
        await writer.drain()
        writer.close()
    return work



# -*- coding: utf-8 -*-

# Import python libs
import os
import importlib
import inspect
import logging
import secrets
import sys

# Import pop libs
import pop.dirs
import pop.scanner
import pop.loader
import pop.exc
import pop.contract

EXT_SUFFIXES = tuple(importlib.machinery.EXTENSION_SUFFIXES)
log = logging.getLogger(__name__)


def ex_path(path):
    '''
    Take a path that is sent to the Sub and expand it if it is a string or not
    '''
    if path is None:
        return []
    elif isinstance(path, str):
        return path.split(',')
    elif isinstance(path, list):
        return path
    return []


class Hub:
    '''
    The redistributed pop central hub. All components of the system are
    rooted to the Hub.
    '''
    def __init__(self):
        self._subs = {}
        self._dynamic = {}
        self._dscan = False
        self._subs['pop'] = Sub(
                self,
                'pop',
                pypath='pop.mods.pop')
        self._iter_subs = sorted(self._subs.keys())
        self._iter_ind = 0

    def __getstate__(self):
        return dict(
            _subs=self._subs,
        )

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __iter__(self):
        def iter(subs):
            for sub in sorted(subs.keys()):
                yield subs[sub]
        return iter(self._subs)

    def _resolve_this(self, levels):
        '''
        This function allows for hub to pop introspective calls.
        This should only ever be called from within a hub module, otherwise
        it should stack trace, or return heaven knows what...
        '''
        if hasattr(sys, '_getframe'):  # implementation detail of CPython, speeds up things by 100x.
            desired_frame = sys._getframe(3)
            contracted = desired_frame.f_locals['self']
        else:
            call_frame = inspect.stack(0)[3]
            contracted = call_frame[0].f_locals['self']
        ref = contracted.ref.split('.')

        # (0=module, 1=module's parent etc.)
        level_offset = levels - 1
        traversed = self
        for i in range(len(ref) - level_offset):
            traversed = getattr(traversed, ref[i])
        return traversed

    def _remove_subsystem(self, subname):
        '''
        Remove the named subsystem
        '''
        if subname in self._subs:
            # Remove the subsystem
            self._subs.pop(subname)
            # reset the iterator
            self._iter_subs = sorted(self._subs.keys())
            self._iter_ind = 0
            return True
        return False

    def _scan_dynamic(self):
        '''
        Refresh the dynamic roots data used for loading app merge module roots
        '''
        self._dynamic = pop.dirs.dynamic_dirs()
        self._dscan = True

    def __getattr__(self, item):
        if item.startswith('_'):
            if item == item[0]*len(item):
                return self._resolve_this(len(item))
            else:
                return self.__getattribute__(item)
        if '.' in item:
            return self.pop.ref.last(item)
        if item in self._subs:
            return self._subs[item]
        return self.__getattribute__(item)


class Sub:
    '''
    The pop object contains the loaded module data
    '''
    def __init__(
            self,
            hub,
            subname,
            pypath=None,
            static=None,
            contracts_pypath=None,
            contracts_static=None,
            default_contracts=None,
            virtual=True,
            dyne_name=None,
            omit_start=('_'),
            omit_end=(),
            omit_func=False,
            omit_class=False,
            omit_vars=False,
            mod_basename='',
            stop_on_failures=False,
            init=True,
            is_contract=False,
            ):
        self._iter_ind = 0
        self._hub = hub
        self._subs = {}
        self._mem = {}
        self._subname = subname
        self._pypath = ex_path(pypath)
        self._static = ex_path(static)
        self._contracts_pypath = ex_path(contracts_pypath)
        self._contracts_static = ex_path(contracts_static)
        if isinstance(default_contracts, str):
            default_contracts = [default_contracts]
        self._default_contracts = default_contracts or ()
        self._dyne_name = dyne_name
        self._virtual = virtual
        self._omit_start = omit_start
        self._omit_end = omit_end
        self._omit_func = omit_func
        self._omit_class = omit_class
        self._omit_vars = omit_vars
        self._mod_basename = mod_basename
        self._stop_on_failures = stop_on_failures
        self._init = init
        self._is_contract = is_contract
        self._prepare()

    def _prepare(self):
        self._dirs = pop.dirs.dir_list(
            self._subname,
            'mods',
            self._pypath,
            self._static,
            )
        self._load_dyne()
        self._contract_dirs = pop.dirs.dir_list(
            self._subname,
            'contracts',
            self._contracts_pypath,
            self._contracts_static,
            )
        self._contract_dirs.extend(
            pop.dirs.inline_dirs(
                self._dirs,
                'contracts'
                )
            )
        if self._contract_dirs:
            self._contracts = Sub(
                self._hub,
                f'{self._subname}.contracts',
                static=self._contract_dirs,
                is_contract=True,
            )
        else:
            self._contracts = None
        self._name_root = self._load_name_root()
        self._mem = {}
        self._scan = pop.scanner.scan(self._dirs)
        self._loaded = {}
        self._vmap = {}
        self._load_errors = {}
        self._loaded_all = False

    def _load_dyne(self):
        '''
        Load up the dynamic dirs for this sub
        '''
        if not self._hub._dscan:
            self._hub._scan_dynamic()
        for path in self._hub._dynamic.get(self._dyne_name, []):
            self._dirs.append(path)

    def _load_name_root(self):
        '''
        Generate the root of the name to be used to apply to the loaded modules
        '''
        if self._pypath:
            return self._pypath[0]
        elif self._dirs:
            return secrets.token_hex()

    def __getstate__(self):
        return dict(
            _hub=self._hub,
            _subname=self._subname,
            _pypath=self._pypath,
            _static=self._static,
            _contracts_pypath=self._contracts_pypath,
            _contracts_static=self._contracts_static,
            _default_contracts=self._default_contracts,
            _virtual=self._virtual,
            _omit_start=self._omit_start,
            _omit_end=self._omit_end,
            _omit_func=self._omit_func,
            _omit_class=self._omit_class,
            _omit_vars=self._omit_vars,
            _mod_basename=self._mod_basename,
            _stop_on_failures=self._stop_on_failures,
            _init=self._init
        )

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._prepare()

    def __getattr__(self, item):
        '''
        If the item should be loaded, load it, else serve it
        '''
        if item.startswith('_'):
            return self.__getattribute__(item)
        if '.' in item:
            return self._hub.pop.ref.last(f'{self._subname}.{item}')
        if item in self._loaded:
            ret = self._loaded[item]
            # If this previously errored on load, try it again,
            # it might be ready to load now
            if isinstance(ret, pop.loader.LoadError):
                ret = self._find_mod(item)
                if isinstance(ret, pop.loader.LoadError):
                    # If this is still a LoadError, process it
                    self._process_load_error(ret)
            return ret
        elif item in self._subs:
            return self._subs[item]
        return self._find_mod(item)

    def __contains__(self, item):
        try:
            return hasattr(self, item)
        except pop.exc.PopLookupError:
            return False

    # @property
    # def __name__(self):
    #     return '{}.{}'.format(self._mod_basename, self._subname)

    def __iter__(self):
        self._load_all()
        def iter(loaded):
            for l in sorted(loaded.keys()):
                yield loaded[l]
        return iter(self._loaded)

    def __next__(self):
        self._load_all()
        if self._iter_ind == len(self._iter_keys):
            self._iter_ind = 0
            raise StopIteration
        self._iter_ind += 1
        return self._loaded[self._iter_keys[self._iter_ind - 1]]

    def _sub_init(self, init):
        '''
        Run load init.py for the sub, running '__init__' function if present
        '''
        if init:
            self._find_mod('init', match_only=True)

    def _process_load_error(self, mod, skip_full_stop=False):
        if not isinstance(mod, pop.loader.LoadError):
            # This is not a LoadError, return now!
            return False

        if mod.edict['verror']:
            error = '{0[msg]}: {0[verror]}'.format(mod())
            if skip_full_stop is False and self._stop_on_failures is True:
                raise pop.exc.PopError(error)
            log.info(error)
            return
        error = '{0[msg]}: {0[exception]!r}'.format(mod())
        if mod.traceback:
            error += '\n' + mod.traceback
        if skip_full_stop is False and self._stop_on_failures is True:
            raise pop.exc.PopError(error)
        if mod.traceback:
            log.warning(error)
        else:
            log.info(error)
        return True

    def _find_mod(self, item, match_only=False):
        '''
        find the module named item
        '''
        for iface in self._scan:
            for bname in self._scan[iface]:
                if os.path.basename(bname) == item:
                    self._load_item(iface, bname)
            if item in self._loaded:
                return self._loaded[item]
        if not match_only:
            for iface in self._scan:
                for bname in self._scan[iface]:
                    if self._scan[iface][bname].get('loaded'):
                        continue
                    self._load_item(iface, bname)
                    if item in self._loaded:
                        return self._loaded[item]
        # Let's see if the module being lookup is in the load errors dictionary
        if item in self._load_errors:
            # Return the LoadError
            return self._load_errors[item]

    def _load_item(self, iface, bname):
        '''
        Load the named basename
        '''
        if iface not in self._scan:
            raise pop.exc.PopLoadError('Bad call to load item, no iface {}'.format(iface))
        if bname not in self._scan[iface]:
            raise pop.exc.PopLoadError(
                'Bad call to load item, no bname {} in iface {}'.format(bname, iface))
        mname = '{}.{}'.format(self._name_root, os.path.basename(bname))
        mod = pop.loader.load_mod(
                mname,
                iface,
                self._scan[iface][bname]['path'],
        )
        if self._process_load_error(mod):
            self._load_errors[os.path.basename(bname)] = mod
            return
        self._prep_mod(mod, iface, bname)

    def _prep_mod(self, mod, iface, bname):
        '''
        Prepare the module!
        '''
        vret = pop.loader.load_virtual(
                self._hub,
                self._virtual,
                mod,
                bname)
        if 'error' in vret:
            # Virtual Errors should not full stop pop
            self._process_load_error(vret['error'], skip_full_stop=True)
            # Store the LoadError under the __virtualname__ if defined
            self._load_errors[vret['vname']] = vret['error']
            return

        contracts = pop.contract.load_contract(
                self._contracts,
                self._default_contracts,
                mod,
                vret['name'])
        name = vret['name']
        if name.endswith(EXT_SUFFIXES):
            for ext in EXT_SUFFIXES:
                if name.endswith(ext):
                    name = name.split(ext)[0]
                    break
        mod_dict = pop.loader.prep_loaded_mod(
                self,
                mod,
                name,
                contracts)
        if name != 'init':
            pop.verify.contract(self._hub, contracts, mod_dict)
        self._loaded[name] = mod_dict
        self._vmap[mod.__file__] = name
        # Let's mark the module as loaded
        self._scan[iface][bname]['loaded'] = True
        # Now that the module has been added to the sub, call mod_init
        pop.loader.mod_init(self._hub, mod)

    def _load_all(self):
        '''
        Load all modules found during the scan.

        .. attention:: This completely disables the lazy loader behavior or pop
        '''
        if self._loaded_all is True:
            return
        for iface in self._scan:
            for bname in self._scan[iface]:
                if self._scan[iface][bname].get('loaded'):
                    continue
                self._load_item(iface, bname)
        self._loaded_all = True



'''
Routines to verify the working environment etc.
'''
# Import python libs
import os


def env(hub):
    '''
    Verify that the directories specified in the system exist
    '''
    for key in hub.opts:
        if key.endswith('_dir'):
            try:
                os.makedirs(hub.opts[key])
            except OSError:
                pass




# -*- coding: utf-8 -*-
'''
Configuration file core loading functions
'''

# Import python libs
import os
import glob
import fnmatch

__virtualname__ = 'file'
__contracts__ = [__virtualname__]


def load_file(hub, paths, defaults=None, overrides=None, includes=True):
    '''
    Load a single configuration file
    '''
    opts = {}
    if isinstance(defaults, dict):
        opts.update(defaults)
    if not isinstance(paths, list):
        paths = paths.split(',')
    add = []
    for fn_ in paths:
        add.extend(glob.glob(fn_))
    paths.extend(add)
    for fn_ in paths:
        if hub.conf._loader == 'yaml':
            opts.update(hub.conf.yaml.load(fn_))
        elif hub.conf._loader == 'json':
            opts.update(hub.conf.json.load(fn_))
        elif hub.conf._loader == 'toml':
            opts.update(hub.conf.toml.load(fn_))
    if includes:
        hub.conf.file.proc_include(opts)
    if isinstance(overrides, dict):
        opts.update(overrides)
    return opts


def load_dir(hub,
             confdir,
             defaults=None,
             overrides=None,
             includes=True,
             recurse=False,
             pattern=None):
    '''
    Load takes a directory location to scan for configuration files. These
    files will be read in. The defaults dict defines what
    configuration options should exist if not found in the confdir. Overrides
    are configuration options which should be included regardless of whether
    those options existed before. If includes is set to True, then the
    statements 'include' and 'include_dir' found in either the defaults or
    in configuration files.
    '''
    opts = {}
    if not isinstance(confdir, list):
        confdir = confdir.split(',')
    confdirs = []
    for dirs in confdir:
        if not isinstance(dirs, (list, tuple)):
            dirs = [dirs]
        for dir_ in dirs:
            confdirs.extend(glob.glob(dir_))
    if isinstance(defaults, dict):
        opts.update(defaults)
    paths = []
    for dir_ in confdirs:
        dirpaths = []
        if os.path.isdir(dir_):
            if not recurse:
                for fn_ in os.listdir(dir_):
                    path = os.path.join(dir_, fn_)
                    if os.path.isdir(path):
                        # Don't process directories
                        continue
                    if pattern and not fnmatch.fnmatch(fn_, pattern):
                        continue
                    dirpaths.append(path)
            else:
                for root, dirs, files in os.walk(dir_):
                    for fn_ in files:
                        path = os.path.join(root, fn_)
                        if pattern and not fnmatch.fnmatch(fn_, pattern):
                            continue
                        dirpaths.append(path)

        # Sort confdir directory paths like:
        # /b.txt
        # /c.txt
        # /a/x.txt
        # /b/x.txt
        paths.extend(sorted(dirpaths, key=lambda p: (p.count(os.path.sep), p)))
    opts.update(hub.conf.file.load_file(paths, includes))
    if isinstance(overrides, dict):
        opts.update(overrides)
    return opts


def proc_include(hub, opts):
    '''
    process include and include_dir
    '''
    rec = False
    if opts.get('include_dir'):
        idir = opts.pop('include_dir')
        opts.update(hub.conf.file.load_dir(idir))
        rec = True
    if opts.get('include'):
        ifn = opts.pop('include')
        opts.update(hub.conf.file.load_file(ifn))
        rec = True
    if rec:
        hub.conf.file.proc_include(opts)
    return opts



# -*- coding: utf-8 -*-
'''
Define the JSON loader interface
'''

# Import python libs
import json

__virtualname__ = 'json'
__contracts__ = [__virtualname__]


def __virtual__(hub):
    return True


def load(hub, path):
    '''
    Use json to read in a file
    '''
    try:
        with open(path, 'r') as fp_:
            ret = json.loads(fp_.read())
        return ret
    except FileNotFoundError:
        return {}


def render(hub, val):
    '''
    Take the string and render it in json
    '''
    return json.loads(val)


# -*- coding: utf-8 -*-
'''
Contracts to enforce loader objects
'''

# Import python libs
import inspect
import os
from collections import namedtuple

# Import pop libs
import pop.exc
import pop.verify


class ContractedContext(namedtuple('ContractedContext', ('func', 'args', 'kwargs', 'signature', 'ret', 'cache'))):
    '''
    Contracted function calling context
    '''
    def __new__(cls, func, args, kwargs, signature, ret=None, cache=None):  # pylint: disable=too-many-arguments
        if cache is None:
            cache = {}
        return super(ContractedContext, cls).__new__(cls, func, list(args), kwargs, signature, ret, cache)

    def get_argument(self, name):
        '''
        Return the value corresponding to a function argument after binding the contract context
        argument and keyword arguments to the function signature.
        '''
        return self.get_arguments()[name]

    def get_arguments(self):
        '''
        Return a dictionary of all arguments that will be passed to the function and their
        values, including default arguments.
        '''
        if '__bound_signature__' not in self.cache:
            try:
                self.cache['__bound_signature__'] = self.signature.bind(*self.args, **self.kwargs)
            except TypeError as e:
                for frame in inspect.trace():
                    if frame.function == 'bind' and frame.filename.endswith(os.sep+'inspect.py'):
                        raise pop.exc.BindError(e)
                raise
            # Apply any default values from the signature
            self.cache['__bound_signature__'].apply_defaults()
        return self.cache['__bound_signature__'].arguments


def load_contract(contracts, default_contracts, mod, name):
    '''
    return a Contract object loaded up
    '''
    raws = []
    if not contracts:
        return raws
    loaded_contracts = []
    if getattr(contracts, name):
        loaded_contracts.append(name)
        raws.append(getattr(contracts, name))
    if getattr(contracts, 'init'):
        loaded_contracts.append('init')
        raws.append(getattr(contracts, 'init'))
    if default_contracts:
        for contract in default_contracts:
            if contract in loaded_contracts:
                continue
            loaded_contracts.append(contract)
            raws.append(getattr(contracts, contract))
    if hasattr(mod, '__contracts__'):
        cnames = getattr(mod, '__contracts__')
        if not isinstance(cnames, (list, tuple)):
            cnames = cnames.split(',')
        for cname in cnames:
            if cname in contracts:
                if cname in loaded_contracts:
                    continue
                loaded_contracts.append(cname)
                raws.append(getattr(contracts, cname))
    return raws


class Wrapper:  # pylint: disable=too-few-public-methods
    def __init__(self, func, ref, name):
        self.func = func
        self.ref = ref
        self.name = name
        self.signature = inspect.signature(self.func)
        self._sig_errors = []

    def __call__(self, *args, **kwargs):
        self.func(*args, **kwargs)

    def __repr__(self):
        return '<{} func={}.{}>'.format(self.__class__.__name__, self.func.__module__, self.name)


class Contracted(Wrapper):  # pylint: disable=too-few-public-methods
    '''
    This class wraps functions that have a contract associated with them
    and executes the contract routines
    '''
    def __init__(self, hub, contracts, func, ref, name):
        super().__init__(func, ref, name)
        self.hub = hub
        self.contracts = contracts if contracts else []
        self._load_contracts()

    def _get_contracts_by_type(self, contract_type='pre'):
        matches = []
        fn_contract_name = '{}_{}'.format(contract_type, self.name)
        for contract in self.contracts:
            if hasattr(contract, fn_contract_name):
                matches.append(getattr(contract, fn_contract_name))
            if hasattr(contract, contract_type):
                matches.append(getattr(contract, contract_type))

        return matches

    def _load_contracts(self):
        self.contract_functions = {'pre': self._get_contracts_by_type('pre'),
                                   'call': self._get_contracts_by_type('call')[:1],
                                   'post': self._get_contracts_by_type('post'),
                                   }
        self._has_contracts = sum([len(l) for l in self.contract_functions.values()]) > 0

    def __call__(self, *args, **kwargs):
        args = (self.hub,) + args

        if not self._has_contracts:
            return self.func(*args, **kwargs)
        contract_context = ContractedContext(self.func, args, kwargs, self.signature)

        for fn in self.contract_functions['pre']:
            fn(contract_context)
        if self.contract_functions['call']:
            ret = self.contract_functions['call'][0](contract_context)
        else:
            ret = self.func(*args, **kwargs)
        for fn in self.contract_functions['post']:
            post_ret = fn(contract_context._replace(ret=ret))
            if post_ret is not None:
                ret = post_ret

        return ret



'''
This module is used to manage the process started up by the pool. Work in this
module is used to manage the worker process itself and not other routines on
the hub this process was derived from

This is an exec, not a fork! This is a fresh memory space!
'''
# Import python libs
import os
import types
import asyncio
# Import third party libs
import msgpack
# TODO: The workers should detect if their controlling process dies and terminate by themselves
# The controlling process will kill them when it exists, but if it exists hard then the workers
# Should be able to also clean themselves up


def start(hub, sock_dir, ind, ref, ret_ref):
    '''
    This function is called by the startup script to create a worker process

    :NOTE: This is a new process started from the shell, it does not have any
    of the process namespace from the creating process.
    This is an EXEC, NOT a FORK!
    '''
    hub.proc.SOCK_DIR = sock_dir
    hub.proc.REF = ref
    hub.proc.SOCK_PATH = os.path.join(sock_dir, ref)
    hub.proc.RET_REF = ret_ref
    hub.proc.RET_SOCK_PATH = os.path.join(sock_dir, ret_ref)
    hub.proc.IND = ind
    hub.pop.loop.start(hub.proc.worker.hold(), hub.proc.worker.server())


async def hold(hub):
    '''
    This function just holds the loop open by sleeping in a while loop
    '''
    while True:
        await asyncio.sleep(60)


async def server(hub):
    '''
    Start the unix socket server to receive commands
    '''
    await asyncio.start_unix_server(
            hub.proc.worker.work,
            path=hub.proc.SOCK_PATH)


async def work(hub, reader, writer):
    '''
    Process the incoming work
    '''
    inbound = await reader.readuntil(hub.proc.DELIM)
    inbound = inbound[:-len(hub.proc.DELIM)]
    payload = msgpack.loads(inbound, encoding='utf8')
    ret = b''
    if 'fun' not in payload:
        ret = {'err': 'Invalid format'}
    elif payload['fun'] == 'sub':
        # Time to add a sub to the hub!
        try:
            hub.proc.worker.add_sub(payload)
            ret = {'status': True}
        except Exception as exc:
            ret = {'status': False, 'exc': str(exc)}
    elif payload['fun'] == 'run':
        # Time to do some work!
        try:
            ret = await hub.proc.worker.run(payload)
        except Exception as exc:
            ret = {'status': False, 'exc': str(exc)}
    elif payload['fun'] == 'gen':
        ret = await hub.proc.worker.gen(payload, reader, writer)
    elif payload['fun'] == 'setattr':
        ret = await hub.proc.worker.set_attr(payload)
    ret = msgpack.dumps(ret, use_bin_type=True)
    ret += hub.proc.D_FLAG
    ret += hub.proc.DELIM
    writer.write(ret)
    await writer.drain()
    writer.close()


def add_sub(hub, payload):
    '''
    Add a new sub onto the hub for this worker
    '''
    hub.pop.sub.add(*payload['args'], **payload['kwargs'])


async def gen(hub, payload, reader, writer):
    '''
    Run a generator and yield back the returns. Supports a generator and an
    async generator
    '''
    ref = payload.get('ref')
    args = payload.get('args', [])
    kwargs = payload.get('kwargs', {})
    ret = hub.pop.ref.last(ref)(*args, **kwargs)
    if isinstance(ret, types.AsyncGeneratorType):
        async for chunk in ret:
            rchunk = msgpack.dumps(chunk, use_bin_type=True)
            rchunk += hub.proc.I_FLAG
            rchunk += hub.proc.DELIM
            writer.write(rchunk)
            await writer.drain()
    elif isinstance(ret, types.GeneratorType):
        for chunk in ret:
            rchunk = msgpack.dumps(chunk, use_bin_type=True)
            rchunk += hub.proc.I_FLAG
            rchunk += hub.proc.DELIM
            writer.write(rchunk)
            await writer.drain()
    elif asyncio.iscoroutine(ret):
        return await ret
    else:
        return ret
    return ''


async def run(hub, payload):
    '''
    Execute the given payload
    '''
    ref = payload.get('ref')
    args = payload.get('args', [])
    kwargs = payload.get('kwargs', {})
    ret = hub.pop.ref.last(ref)(*args, **kwargs)
    if asyncio.iscoroutine(ret):
        return await ret
    return ret


async def set_attr(hub, payload):
    '''
    Set the named attribute to the hub
    '''
    ref = payload.get('ref')
    value = payload.get('value')
    hub.pop.ref.create(ref, value)


async def ret(hub, payload):
    '''
    Send a return payload to the spawning process. This return will be tagged
    with the index of the process that returned it
    '''
    payload = {'ind': hub.proc.IND, 'payload': payload}
    mp = msgpack.dumps(payload, use_bin_type=True)
    mp += hub.proc.DELIM
    reader, writer = await asyncio.open_unix_connection(path=hub.proc.RET_SOCK_PATH)
    writer.write(mp)
    await writer.drain()
    ret = await reader.readuntil(hub.proc.DELIM)
    ret = ret[:-len(hub.proc.DELIM)]
    writer.close()
    return msgpack.loads(ret, encoding='utf8')



# -*- coding: utf-8 -*-
'''
Define the yaml loader interface
'''

# Import third party libs
try:
    import toml
    HAS_TOML = True
except ImportError:
    HAS_TOML = False

__virtualname__ = 'toml'
#__contracts__ = [__virtualname__]


def __virtual__(hub):
    if HAS_TOML:
        return True
    return (False, 'TOML could not be loaded')


def load(hub, path):
    '''
    use toml to read in a file
    '''
    try:
        with open(path, 'rb') as fp_:
            return toml.load(fp_.read())
    except FileNotFoundError:
        pass
    return {}

def render(hub, val):
    '''
    Take the string and render it in json
    '''
    return toml.loads(val)



'''
This sub is used to set up logging for pop projects and injects logging
options into conf making it easy to add robust logging
'''
# Import python libs
import logging


def __init__(hub):
    '''
    Set up variables used by the log subsystem
    '''
    hub.conf.log.LEVELS = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'critical': logging.CRITICAL,
        }


def conf(hub, name):
    '''
    Return the conf dict for logging, this should be merged OVER by the loaded
    config dict(s)
    '''
    #TODO: Make this more robust to handle more logging interfaces
    ldict = {
        'log_file':
            {
            'default': f'{name}.log',
            'help': 'The location of the log file',
            },
        'log_level':
            {
            'default': 'info',
            'help': 'Set the log level, either quiet, info, warning, or error',
            },
        'log_fmt_logfile':
            {
            'default': '%(asctime)s,%(msecs)03d [%(name)-17s][%(levelname)-8s] %(message)s',
            'help': 'The format to be given to log file messages',
            },
        'log_fmt_console':
            {
            'default': '[%(levelname)-8s] %(message)s',
            'help': 'The log formatting used in the console',
            },
        'log_datefmt':
            {
            'default': '%H:%M:%S',
            'help': 'The date format to display in the logs',
            },
        'log_plugin':
            {
            'default': 'basic',
            'help': 'The logging plugin to use',
            },
        }
    return ldict